import simple_websocket, threading

from ..execution import Execution
from ...contract import dashes_contract
from ..live_execution import LiveExecution


class DashExecution(LiveExecution):
    def _run(self):
        from .dash_runtime import DashRuntime

        self.thread_id = threading.get_ident()
        Execution.executions[self.thread_id] = self

        msg_handler = DashRuntime(self, dash_json=self.runtime_json)  # type: ignore
        self.send(dashes_contract.ExecutionIdMessage(self.id))
        while True:
            try:
                type, data = self.recv()
                msg_handler.handle(type, data)
            except simple_websocket.ConnectionError:
                print("connection closed")



from dataclasses import dataclass
from typing import List, Optional
from pathlib import Path
from tempfile import mkdtemp
from shutil import move

from .json.classes import AbstraJSONRepository
from ..settings import Settings
from importlib_metadata import packages_distributions
from pkg_resources import get_distribution
import ast


@dataclass
class Requirement:
    name: str
    version: Optional[str] = None

    def to_text(self):
        if self.version is None:
            return self.name
        else:
            return f"{self.name}=={self.version}"

    @staticmethod
    def from_text(text: str):
        if "==" in text:
            name, version = text.split("==")
            return Requirement(name=name, version=version)
        else:
            return Requirement(name=text)

    def to_dict(self):
        return {"name": self.name, "version": self.version}

    @staticmethod
    def from_dict(data: dict):
        return Requirement(name=data["name"], version=data["version"])


@dataclass
class Requirements:
    libraries: List[Requirement]

    def to_text(self):
        return "\n".join([lib.to_text() for lib in self.libraries])

    @staticmethod
    def from_text(text: str):
        libraries = []
        for line in text.splitlines():
            if line.startswith("#"):
                continue
            if line.strip() == "":
                continue
            libraries.append(Requirement.from_text(line))
        return Requirements(libraries=libraries)

    def to_dict(self):
        return [lib.to_dict() for lib in self.libraries]

    @staticmethod
    def from_dict(data: list):
        return Requirements(libraries=[Requirement.from_dict(lib) for lib in data])

    def add(self, name: str, version: Optional[str] = None):
        self.libraries.append(Requirement(name=name, version=version))

    def delete(self, name: str):
        self.libraries = [lib for lib in self.libraries if lib.name != name]


@dataclass
class RequirementsRepository:
    @classmethod
    def get_file_path(cls):
        return Settings.root_path / "requirements.txt"

    @classmethod
    def get_recommendation(cls) -> List[Requirement]:
        imported_modules = set()

        abstra_json = AbstraJSONRepository.load()

        for python_file in abstra_json.project_files:
            try:
                code = python_file.read_text(encoding="utf-8")
                parsed = ast.parse(code)
                for node in parsed.body:
                    if isinstance(node, ast.Import):
                        for alias in node.names:
                            lib_name = packages_distributions().get(alias.name)
                            if lib_name is None:
                                continue
                            for l in lib_name:
                                imported_modules.add(l)
                    elif isinstance(node, ast.ImportFrom):
                        if node.module is None:
                            continue
                        lib_name = packages_distributions().get(node.module)
                        if lib_name is None:
                            continue
                        for l in lib_name:
                            imported_modules.add(l)
            except SyntaxError:
                continue

        already_added = set([lib.name for lib in cls.load().libraries])

        return [
            Requirement(name=module, version=get_distribution(module).version)
            for module in imported_modules
            if module not in already_added
        ]

    @classmethod
    def save(cls, requirements: Requirements):
        temp_file = Path(mkdtemp()) / "requirements.txt"

        with temp_file.open("w") as f:
            f.write(requirements.to_text())
        move(str(temp_file), cls.get_file_path())

    @classmethod
    def load(cls) -> Requirements:
        file = cls.get_file_path()
        if file.exists():
            requirements_content = file.read_text(encoding="utf-8")
            return Requirements.from_text(requirements_content)
        else:
            return Requirements(libraries=[])



import flask
from ...execution.execution import RequestData
from ..controller import MainController
from ...usage import usage
from ...settings import Settings
from .utils import send_from_dist
from ...execution.hook_execution import HookExecution
from ...execution.job_execution import JobExecution
from ...execution.script_execution import ScriptExecution
from ...repositories.json.classes import AbstraJSONRepository


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor", __name__)

    @bp.route("/", methods=["GET"])
    @usage
    def _spa_index():
        return send_from_dist("editor.html", "editor.html")

    @bp.route("/api/assets/<path:path>", methods=["GET"])
    def _asset(path):
        return send_from_dist(path, dist_folder=Settings.root_path)

    @bp.route("/api/workspace", methods=["GET"])
    @usage
    def _get_workspace():
        return controller.get_workspace().editor_dto

    @bp.route("/api/workspace", methods=["PUT"])
    @usage
    def _update_workspace():
        if not flask.request.json:
            flask.abort(400)
        workspace = controller.update_workspace(flask.request.json)
        return workspace.editor_dto

    @bp.route("/api/workspace/root-path", methods=["GET"])
    @usage
    def _get_workspace_root_path():
        return str(Settings.root_path.absolute())

    @bp.route("/api/workspace/open-file", methods=["POST"])
    @usage
    def _open_file():
        if not flask.request.json:
            flask.abort(400)
        file_path = flask.request.json["path"]
        controller.open_file(file_path, create_if_not_exists=True)
        return {"success": True}

    @bp.route("/api/workspace/check-file", methods=["GET"])
    @usage
    def _check_file():
        file_path = flask.request.args["path"]
        return {"exists": controller.check_file(file_path)}

    @bp.route("/api/workspace/deploy", methods=["POST"])
    @usage
    def _deploy():
        controller.deploy()
        return {"success": True}

    @bp.route("/api/forms/", methods=["GET"])
    @usage
    def _get_forms():
        return [f.editor_dto for f in controller.get_forms()]

    @bp.route("/api/forms/<path:path>", methods=["GET"])
    @usage
    def _get_form(path: str):
        form = controller.get_form(path)
        if not form:
            flask.abort(404)
        return form.editor_dto

    @bp.route("/api/forms/<path:path>", methods=["DELETE"])
    @usage
    def _delete_form(path: str):
        controller.delete_form(path)
        return {"success": True}

    @bp.route("/api/forms/", methods=["POST"])
    @usage
    def _create_form():
        form = controller.create_form()
        return form.editor_dto

    @bp.route("/api/forms/<path:path>", methods=["PUT"])
    @usage
    def _update_form(path: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        form = controller.update_runtime(path, data)
        return form.editor_dto if form else None

    @bp.route("/api/dashes/<path:path>", methods=["GET"])
    @usage
    def _get_dash(path: str):
        dash = controller.get_dash(path)
        if not dash:
            flask.abort(404)
        return dash.editor_dto

    @bp.route("/api/dashes/", methods=["GET"])
    @usage
    def _get_dashes():
        return [f.editor_dto for f in controller.get_dashes()]

    @bp.route("/api/dashes/", methods=["POST"])
    @usage
    def _create_dash():
        dash = controller.create_dash()
        return dash.editor_dto

    @bp.route("/api/dashes/<path:path>", methods=["PUT"])
    @usage
    def _update_dash(path: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        dash = controller.update_runtime(path, data)
        return dash.editor_dto if dash else None

    @bp.route("/api/dashes/<path:path>", methods=["DELETE"])
    @usage
    def _delete_dash(path: str):
        controller.delete_dash(path)
        return {"success": True}

    @bp.route("/api/hooks/<path:path>", methods=["GET"])
    @usage
    def _get_hook(path: str):
        hook = controller.get_hook(path)
        if not hook:
            flask.abort(404)
        return hook.editor_dto

    @bp.route("/api/hooks/", methods=["GET"])
    @usage
    def _get_hooks():
        return [f.editor_dto for f in controller.get_hooks()]

    @bp.route("/api/hooks/", methods=["POST"])
    @usage
    def _create_hook():
        hook = controller.create_hook()
        return hook.editor_dto

    @bp.route("/api/hooks/<path:path>", methods=["PUT"])
    @usage
    def _update_hook(path: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        hook = controller.update_runtime(path, data)
        return hook.editor_dto if hook else None

    @bp.route("/api/hooks/<path:path>", methods=["DELETE"])
    @usage
    def _delete_hook(path: str):
        controller.delete_hook(path)
        return {"success": True}

    @bp.route(
        "/api/hooks/<path:path>/test", methods=["POST", "GET", "PUT", "DELETE", "PATCH"]
    )
    @usage
    def _test_hook(path: str):
        hook = controller.get_hook(path)
        if not hook:
            flask.abort(404)

        request_data = RequestData(
            method=flask.request.method,
            body=flask.request.get_data(as_text=True),
            headers=flask.request.headers,
            query_params=flask.request.args,
        )

        abstra_json = AbstraJSONRepository.load()
        is_initial = abstra_json.is_initial(hook.path)
        execution = HookExecution(hook, is_initial, request_data)

        execution.run_sync()

        body, status, headers = execution.get_response()

        controller.run_waiting_scripts(execution.stage_run)

        return {
            "body": body,
            "status": status,
            "headers": headers,
            "stdout": "".join(execution.stdout if execution else []),
            "stderr": "".join(execution.stderr if execution else []),
        }

    @bp.route("/api/jobs/<path:identifier>", methods=["GET"])
    @usage
    def _get_job(identifier: str):
        job = controller.get_job(identifier)
        if not job:
            flask.abort(404)
        return job.editor_dto

    @bp.route("/api/jobs/", methods=["GET"])
    @usage
    def _get_jobs():
        return [f.editor_dto for f in controller.get_jobs()]

    @bp.route("/api/jobs/", methods=["POST"])
    @usage
    def _create_job():
        job = controller.create_job()
        return job.editor_dto

    @bp.route("/api/jobs/<path:identifier>", methods=["PUT"])
    @usage
    def _update_runtime(identifier: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        job = controller.update_runtime(identifier, data)
        return job.editor_dto if job else None

    @bp.route("/api/jobs/<path:identifier>", methods=["DELETE"])
    @usage
    def _delete_job(identifier: str):
        controller.delete_job(identifier)
        return {"success": True}

    @bp.route("/api/jobs/<path:identifier>/test", methods=["POST"])
    @usage
    def _test_job(identifier: str):
        job = controller.get_job(identifier)
        if not job:
            flask.abort(404)

        request_data = RequestData(
            method=flask.request.method,
            body=flask.request.get_data(as_text=True),
            headers=flask.request.headers,
            query_params=flask.request.args,
        )

        abstra_json = AbstraJSONRepository.load()
        is_initial = abstra_json.is_initial(job.path)
        execution = JobExecution(job, is_initial, request_data)

        execution.run_sync()

        controller.run_waiting_scripts(execution.stage_run)

        return {
            "stdout": "".join(execution.stdout if execution else []),
            "stderr": "".join(execution.stderr if execution else []),
        }

    @bp.route("/api/scripts/<path:identifier>", methods=["GET"])
    @usage
    def _get_script(identifier: str):
        script = controller.get_script(identifier)
        if not script:
            flask.abort(404)
        return script.editor_dto

    @bp.route("/api/scripts/", methods=["GET"])
    @usage
    def _get_scripts():
        return [f.editor_dto for f in controller.get_scripts()]

    @bp.route("/api/scripts/", methods=["POST"])
    @usage
    def _create_script():
        script = controller.create_script()
        return script.editor_dto

    @bp.route("/api/scripts/<path:identifier>", methods=["PUT"])
    @usage
    def _update_script(identifier: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        script = controller.update_runtime(identifier, data)
        return script.editor_dto if script else None

    @bp.route("/api/scripts/<path:identifier>", methods=["DELETE"])
    @usage
    def _delete_script(identifier: str):
        controller.delete_script(identifier)
        return {"success": True}

    @bp.route("/api/scripts/<path:path>/test", methods=["POST"])
    @usage
    def _test_script(path: str):
        script = controller.get_script(path)

        if not script:
            flask.abort(404)

        return controller.run_initial_script(script)

    @bp.route("/api/stage_runs", methods=["GET"])
    @usage
    def _get_stage_runs():
        return controller.get_stage_runs()

    @bp.route("/api/workflow-editor/initial-data", methods=["GET"])
    @usage
    def _workflow_initial_data():
        try:
            return controller.workflow_initial_data()
        except Exception as e:
            return str(e), 500

    @bp.route("/api/workflow-editor/move", methods=["POST"])
    @usage
    def _workflow_move():
        try:
            payload = flask.request.json
            controller.workflow_move(payload)
            return flask.Response(status=204)
        except Exception as e:
            return str(e), 500

    @bp.route("/api/workflow-editor/add-nodes", methods=["POST"])
    @usage
    def _workflow_add_nodes():
        try:
            payload = flask.request.json
            controller.workflow_add_nodes(payload)
            return flask.Response(status=204)
        except Exception as e:
            return str(e), 500

    @bp.route("/api/workflow-editor/duplicate-nodes", methods=["POST"])
    @usage
    def _workflow_duplicate_nodes():
        try:
            payload = flask.request.json
            controller.workflow_duplicate_nodes(payload)
            return flask.Response(status=204)
        except Exception as e:
            return str(e), 500

    @bp.route("/api/workflow-editor/delete", methods=["POST"])
    @usage
    def _workflow_delete():
        try:
            payload = flask.request.json
            controller.workflow_delete(payload)
            return flask.Response(status=204)
        except Exception as e:
            return str(e), 500

    @bp.route("/api/workflow-editor/add-transition", methods=["POST"])
    @usage
    def _workflow_add_transition():
        try:
            payload = flask.request.json
            controller.workflow_add_transition(payload)
            return flask.Response(status=204)
        except Exception as e:
            return str(e), 500

    @bp.route("/api/requirements", methods=["GET"])
    @usage
    def _get_requirements():
        return controller.get_requirements()

    @bp.route("/api/requirements", methods=["POST"])
    @usage
    def _create_requirement():
        data = flask.request.json

        if not data:
            flask.abort(400)

        return controller.create_requirement(data["name"], data.get("version"))

    @bp.route("/api/requirements/<name>", methods=["DELETE"])
    @usage
    def _delete_requirement(name: str):
        return controller.delete_requirement(name)

    @bp.route("/api/requirements-recommendations", methods=["GET"])
    @usage
    def _get_requirements_recommendation():
        return controller.get_requirements_recommendations()

    @bp.route("/api/debugger", methods=["GET"])
    @usage
    def _get_debugger_status():
        return controller.get_debugger_status()

    @bp.route("/api/debugger/vscode-launch", methods=["POST"])
    @usage
    def _create_vscode_launch():
        return controller.create_vscode_launch()

    @bp.route("/api/login", methods=["GET"])
    @usage
    def _get_login():
        return controller.get_login()

    @bp.route("/api/login", methods=["POST"])
    @usage
    def _create_login():
        data = flask.request.json
        if not data:
            flask.abort(400)

        return controller.create_login(data["token"])

    @bp.route("/api/login", methods=["DELETE"])
    @usage
    def _delete_login():
        return controller.delete_login()

    @bp.route("/api/project-info", methods=["GET"])
    def _get_project_info():
        return controller.get_project_info()

    @bp.route("/api/ai/message", methods=["POST"])
    @usage
    def _get_next_message():
        body = flask.request.json
        if not body:
            flask.abort(400)

        streamer = controller.send_ai_message(body["messages"], body["runtime"])

        if streamer is None:
            flask.abort(403)

        return flask.Response(streamer, mimetype="text/event-stream")

    @bp.route("/<path:filename>", methods=["GET"])
    @usage
    def _spa(filename: str):
        return send_from_dist(filename, "editor.html")

    return bp



from .generated.inputs import *
from .generated.outputs import *
from .page import Page  # exported
from .step import run_steps  # exported
from .list_item_schema import ListItemSchema  # exported
from ..common import execute_js, redirect, get_user, get_query_params, alert  # exported

from abstra_internals.proxy import ReadOnlyProxyDict

# legacy
url_params = ReadOnlyProxyDict(get_query_params)



import flask, flask_sock

from ..controller import MainController
from .auth import get_auth_bp
from ...settings import Settings
from .utils import send_from_dist
from ...execution.execution import RequestData
from ...utils.environment import BUILD_ID, SIDECAR_SHARED_TOKEN
from ...execution import HookExecution, JobExecution, DashExecution, FormExecution
from ...repositories.json.classes import AbstraJSONRepository


def get_player_bp(controller: MainController):
    bp = flask.Blueprint("player", __name__)
    sock = flask_sock.Sock(bp)

    auth_bp = get_auth_bp(controller)
    bp.register_blueprint(auth_bp, url_prefix="/_auth")

    @bp.route("/_pages/<path:id_or_path>", methods=["GET"])
    def get_runner_data(id_or_path):
        runtime = controller.get_page_runtime(id_or_path)
        if not runtime:
            print("404", id_or_path)
            flask.abort(404)

        workspace = controller.get_workspace()
        res = {}
        res[runtime.runner_type] = {
            **runtime.browser_runner_dto,
            "workspace": workspace.browser_runner_dto,  # TODO: fix this in frontend
        }
        return res

    @bp.route("/_version", methods=["GET"])
    def _get_version():
        return BUILD_ID

    @sock.route("/_socket")
    def websocket(conn: flask_sock.Server):
        request_data = RequestData(
            query_params=flask.request.args,
            body=flask.request.get_data(as_text=True),
            headers=flask.request.headers,
            method=flask.request.method,
        )

        try:
            dash_path = flask.request.args.get("dashPath")
            if dash_path is not None:
                dash = controller.get_dash(dash_path)

                abstra_json = AbstraJSONRepository.load()
                is_initial = abstra_json.is_initial(dash_path)

                if not dash:
                    conn.close(reason=404, message="Not found")
                    return

                return DashExecution(dash, is_initial, conn, request_data).run_sync()

            form_path = flask.request.args.get("formPath")
            if form_path is not None:
                form = controller.get_form(form_path)

                abstra_json = AbstraJSONRepository.load()
                is_initial = abstra_json.is_initial(form_path)

                if not form:
                    conn.close(reason=404, message="Not found")
                    return

                execution = FormExecution(form, is_initial, conn, request_data)

                execution.run_sync()

                controller.run_waiting_scripts(execution.stage_run)

        finally:
            conn.close(message="Done")

    @bp.route("/_files", methods=["PUT"])
    def _upload_file():
        files = flask.request.files
        filename = flask.request.form.get("filename")
        if filename is None:
            flask.abort(400)

        if len(files) == 0:
            flask.abort(400)

        paths = []
        for file in files.values():
            paths.append(controller.save_file(file, filename))
        return paths

    @bp.route("/_files/<path:path>", methods=["GET"])
    def _get_file(path):
        return flask.send_file(controller.get_file(path))

    @bp.route("/_assets/logo", methods=["GET"])
    def _logo():
        logo_path = controller.get_workspace().logo_url
        if not logo_path:
            return flask.abort(404)
        return send_from_dist(logo_path, dist_folder=Settings.root_path)

    @bp.route("/favicon.ico", methods=["GET"])
    def _favicon():
        return _logo()

    @bp.route("/_assets/background", methods=["GET"])
    def _background():
        background_path = controller.get_workspace().theme

        if not background_path:
            return flask.abort(404)

        return send_from_dist(background_path, dist_folder=Settings.root_path)

    @bp.route("/_hooks/<path:path>", methods=["POST", "GET", "PUT", "DELETE", "PATCH"])
    def hook_runner(path):
        hook = controller.get_hook(path)
        if not hook:
            flask.abort(404)

        if not hook.file:
            flask.abort(500)

        request_data = RequestData(
            query_params=flask.request.args,
            body=flask.request.get_data(as_text=True),
            headers=flask.request.headers,
            method=flask.request.method,
        )

        abstra_json = AbstraJSONRepository.load()
        is_initial = abstra_json.is_initial(path)

        execution = HookExecution(hook, is_initial, request_data)

        execution.run_sync()

        controller.run_waiting_scripts(execution.stage_run)

        body, status, headers = execution.get_response()

        return flask.Response(status=status, headers=headers, response=body)

    @bp.route("/_jobs/<path:path>", methods=["POST"])
    def job_runner(path):
        if flask.request.headers.get("Shared-Token") != SIDECAR_SHARED_TOKEN:
            flask.abort(401)

        job = controller.get_job(path)
        if not job:
            flask.abort(404)

        if not job.file:
            flask.abort(500)

        request_data = RequestData(
            method=flask.request.method,
            body=flask.request.get_data(as_text=True),
            headers=flask.request.headers,
            query_params=flask.request.args,
        )

        def run_job(job):
            abstra_json = AbstraJSONRepository.load()
            is_initial = abstra_json.is_initial(path)

            execution = JobExecution(job, is_initial, request_data)

            execution.run_sync()
            controller.run_waiting_scripts(execution.stage_run)

        controller.executor.submit(run_job, job)

        return {"status": "running"}

    @bp.route("/<path:filename>", methods=["GET"])
    def spa(filename: str):
        res = send_from_dist(filename, "player.html")
        return res

    return bp



import flask_sock, typing, copy

from ..repositories.stage_run import StageRun
from ..repositories import StageRunRepository
from .live_execution import LiveExecution
from ..contract import forms_contract
from .execution import RequestData


class FormExecution(LiveExecution):
    def _wait_start(self):
        type = None
        data = None

        type, data = self.recv()

        if type != "start":
            return self._wait_start()

        return data["params"]

    def _handle_abandoned_middle_form(
        self, initial_abandoned_stage_run: StageRun, parent_id: str
    ):
        child_runs = StageRunRepository.find({"parent_id": parent_id})

        assert len(child_runs) == 1, "Expected 1 child run for abandoned stage run"
        child_run = child_runs[0]

        assert (
            child_run.stage == initial_abandoned_stage_run.stage
        ), "Expected child run to have same stage as initial abandoned stage run"
        assert (
            child_run.assignee == initial_abandoned_stage_run.assignee
        ), "Expected child run to have same assignee as initial abandoned stage run"

        if child_run.is_abandoned:
            return self._handle_abandoned_middle_form(
                initial_abandoned_stage_run, child_run.id
            )

        self.stage_run = child_run

    def init_stage_run(self, id: typing.Optional[str] = None):
        super().init_stage_run(id)
        if self.stage_run and not self.is_initial and self.stage_run.is_abandoned:
            return self._handle_abandoned_middle_form(self.stage_run, self.stage_run.id)

    def setup_context(self, request: RequestData):
        self.query_params = self._wait_start()
        self.send(forms_contract.ExecutionIdMessage(self.id))
        self.init_stage_run(self.query_params.get(self.abstra_run_key))

    def handle_success(self) -> str:
        close_dto = forms_contract.CloseDTO(exit_status="SUCCESS")
        self.send(forms_contract.CloseMessage(close_dto))
        return super().handle_success()

    def _handle_ws_exception_1001(self, exception: flask_sock.ConnectionClosed) -> str:
        self.log(
            "connection-closed",
            {"message": "Client went away - probably closed the tab."},
        )

        if self.stage_run and self.stage_run_freezed and not self.is_initial:
            StageRunRepository.create_next(
                self.stage_run,
                [
                    dict(
                        assignee=self.stage_run_freezed.assignee,
                        stage=self.stage_run_freezed.stage,
                        data=self.stage_run.data,
                    )
                ],
            )

        return "abandoned"

    def _handle_ws_exception_other(self, exception: flask_sock.ConnectionClosed) -> str:
        self.log(
            "connection-closed",
            {
                "message": f"[ERROR] Connection closed with code {exception.reason}: {exception.message}\n",
                "reason": exception.reason,
            },
        )
        return super().handle_failure(exception)

    def _handle_ws_exception(self, exception: flask_sock.ConnectionClosed) -> str:
        if exception.reason == 1000:
            return super().handle_success()  # missing advanve steps?

        if exception.reason == 1001:
            return self._handle_ws_exception_1001(exception)

        return self._handle_ws_exception_other(exception)

    def handle_failure(self, exception: Exception) -> str:
        if isinstance(exception, flask_sock.ConnectionClosed):
            return self._handle_ws_exception(exception)

        close_dto = forms_contract.CloseDTO(
            exit_status="GENERIC_EXCEPTION",
            exception=exception.__str__(),
        )
        self.send(forms_contract.CloseMessage(close_dto))

        return super().handle_failure(exception)

    def handle_finish(self):
        self.close()



from .ClickInput import ClickInput
from .CodeInput import CodeInput
from .ListInput import ListInput
from .KanbanBoardInput import KanbanBoardInput
from .CnpjInput import CnpjInput
from .TextareaInput import TextareaInput
from .DropdownInput import DropdownInput
from .CustomInput import CustomInput
from .EmailInput import EmailInput
from .LinkOutput import LinkOutput
from .NumberInput import NumberInput
from .PandasRowSelectionInput import PandasRowSelectionInput
from .FileOutput import FileOutput
from .ImageInput import ImageInput
from .HtmlOutput import HtmlOutput
from .TagInput import TagInput
from .CardsInput import CardsInput
from .TextInput import TextInput
from .RichTextInput import RichTextInput
from .PandasOutput import PandasOutput
from .LatexOutput import LatexOutput
from .NpsInput import NpsInput
from .RatingInput import RatingInput
from .TextOutput import TextOutput
from .CheckboxInput import CheckboxInput
from .CurrencyInput import CurrencyInput
from .ChecklistInput import ChecklistInput
from .ImageOutput import ImageOutput
from .MultipleChoiceInput import MultipleChoiceInput
from .TimeInput import TimeInput
from .DateInput import DateInput
from .PasswordInput import PasswordInput
from .IframeOutput import IframeOutput
from .MarkdownOutput import MarkdownOutput
from .ProgressOutput import ProgressOutput
from .VideoInput import VideoInput
from .PlotlyOutput import PlotlyOutput
from .ToggleInput import ToggleInput
from .CpfInput import CpfInput
from .FileInput import FileInput
from .NumberSliderInput import NumberSliderInput
from .PhoneInput import PhoneInput



import ast
import fnmatch, re
import os
from pathlib import Path
from typing import Generator, Union


GIT_FOLDER = ".git"
ABSTRA_FOLDER = ".abstra"
ABSTRA_IGNORE = ".abstraignore"
CREDENTIALS_FILE = ".abstra/credentials"


def get_ignore_files(dir: Path):
    GITIGNORE = dir.joinpath(".gitignore")
    IGNOREFILE = dir.joinpath(ABSTRA_IGNORE)
    git_path = dir.joinpath(GIT_FOLDER)
    abstra_path = dir.joinpath(ABSTRA_FOLDER)
    ignored = [IGNOREFILE.name, GITIGNORE.name, abstra_path.name, git_path.name]

    if IGNOREFILE.exists():
        with open(IGNOREFILE, "r", encoding="utf-8") as f:
            ignored.extend(IGNOREFILE.read_text(encoding="utf-8").split("\n"))
    elif GITIGNORE.exists():
        ignored.extend(GITIGNORE.read_text(encoding="utf-8").split("\n"))
    return [
        re.compile(fnmatch.translate(f"{p}*"))
        for p in ignored
        if p and not p.startswith("#") and not p.startswith("!")
    ]


def should_ignore(ignored_paths, _path: Path):
    path = str(_path)
    for ignored_path in ignored_paths:
        if ignored_path.search(path):
            return True
    return False


def files_from_directory(directory):
    ignored = [*get_ignore_files(directory), *get_ignore_files(Path.cwd())]
    paths = filter(
        lambda p: p.is_file() and not should_ignore(ignored, p.relative_to(directory)),
        Path(directory).rglob("*"),
    )
    return list(paths)


def path2module(path: Path) -> str:
    if path.suffix == ".py":
        path = path.with_suffix("")
    if path.name == "__init__":
        path = path.parent
    return str(path).replace("/", ".").replace("\\", ".")


def module2path(module: str, package: bool) -> Path:
    path_list = module.split(".")

    # Handle trailing dots
    is_first_trailing_dot = True
    for index, path in enumerate(path_list):
        if path.__len__() == 0:
            if is_first_trailing_dot:
                path_list[index] = "."
                is_first_trailing_dot = False
            else:
                path_list[index] = ".."
        else:
            break

    if package:
        path_list.append("__init__.py")
    else:
        path_list[-1] += ".py"
    return Path(*path_list)


def _get_file_path_from_absolute_module(module: str) -> Union[Path, None]:
    module_path = Path(module2path(module, package=False))
    if module_path.is_file():
        return module_path

    package_path = Path(module2path(module, package=True))
    if package_path.is_file():
        return package_path

    return None


def _get_file_path_from_relative_module(
    module: str, parent_path: Path
) -> Union[Path, None]:
    module_path = parent_path / module2path(module, package=False)
    resolved_module_path = module_path.resolve().relative_to(os.getcwd())

    if resolved_module_path.is_file():
        return resolved_module_path

    package_path = parent_path / module2path(module, package=True)
    resolved_package_path = package_path.resolve().relative_to(os.getcwd())

    if resolved_package_path.is_file():
        return resolved_package_path

    return None


def traverse_code(path: Path) -> Generator[Path, None, None]:
    yield path

    try:
        code = path.read_text(encoding="utf-8")
        parsed = ast.parse(code)

        for node in parsed.body:
            if isinstance(node, ast.Import):
                for alias in node.names:
                    file_path = _get_file_path_from_absolute_module(alias.name)

                    if file_path is not None:
                        yield from traverse_code(file_path)

            if isinstance(node, ast.ImportFrom):
                prefix = "." * (node.level)
                module_name = f"{node.module}." if node.module is not None else ""

                for alias in node.names:
                    relative_module = f"{prefix}{module_name}{alias.name}"

                    file_path = (
                        _get_file_path_from_relative_module(
                            relative_module, path.parent
                        )
                        if node.level > 0
                        else _get_file_path_from_absolute_module(relative_module)
                    )

                    if file_path is not None:
                        yield from traverse_code(file_path)

    except:
        pass



from ..widget_base import Output


class ProgressOutput(Output):
    type = "progress-output"

    def __init__(self, current: float = 50, total: float = 100, **kwargs):
        self.set_props(dict(current=current, total=total, **kwargs))

    def set_props(self, props):
        self.current = props.get("current", 0)
        self.total = props.get("total", 0)
        self.text = props.get("text", "")
        self.full_width = props.get("full_width", False)

    def render(self, context: dict):
        return {
            "type": self.type,
            "current": self.current,
            "total": self.total,
            "text": self.text,
            "fullWidth": self.full_width,
        }



from .static_execution import StaticExecution
from .execution import RequestData


class JobExecution(StaticExecution):
    def setup_context(self, request: RequestData):
        self.init_stage_run()

    def handle_started(self):
        self.log("started", {})
        return super().handle_started()

    def handle_success(self) -> str:
        self.log("success", {})
        return super().handle_success()

    def handle_failure(self, e: Exception) -> str:
        self.log("failed", {"error": str(e)})
        return super().handle_failure(e)



from ..widget_base import Input
from typing import Optional, Union
from ..response_types import PhoneResponse
import re


class PhoneInput(Input):
    type = "phone-input"
    empty_value: PhoneResponse = PhoneResponse(
        raw="", masked="", country_code="", national_number=""
    )

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get("label", "Label")
        self.value = props.get("initial_value", None)
        self.required = props.get("required", True)
        self.hint = props.get("hint", None)
        self.placeholder = props.get("placeholder", "")
        self.full_width = props.get("full_width", False)
        self.disabled = props.get("disabled", False)
        self.invalid_message = props.get(
            "invalid_message", "i18n_error_invalid_phone_number"
        )

    def render(self, context: dict):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "value": self.serialize_value(),
            "placeholder": self.placeholder,
            "required": self.required,
            "hint": self.hint,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
            "errors": self.errors,
            "invalidMessage": self.invalid_message,
        }

    def validate(self):
        errors = []
        if self.required and (
            self.value.country_code == "" or self.value.national_number == ""
        ):
            errors.append("i18n_error_required_field")
        return errors

    def is_value_unset(self):
        return self.value.raw == ""

    def set_value(self, value, set_errors=False):
        if isinstance(value, PhoneResponse):
            self.value = value
        elif isinstance(value, dict):
            self.value = self.parse_value(value)
        elif value == None:
            self.value = self.empty_value

    def serialize_value(self):
        if isinstance(self.value, dict):
            return {
                "countryCode": re.sub(
                    "\\D", "", str(self.value.get("country_code", ""))
                ),
                "nationalNumber": re.sub(
                    "\\D", "", str(self.value.get("national_number", ""))
                ),
            }
        if isinstance(self.value, PhoneResponse):
            return {
                "countryCode": re.sub("\\D", "", str(self.value.country_code)),
                "nationalNumber": re.sub("\\D", "", str(self.value.national_number)),
            }
        return {"countryCode": "", "nationalNumber": ""}

    def parse_value(self, value) -> PhoneResponse:
        if value is None:
            return self.empty_value
        masked = f"+{value.get('countryCode')} {value.get('nationalNumber')}"
        raw = re.sub("\\D", "", masked)
        return PhoneResponse(
            raw=raw,
            masked=masked,
            country_code=f"+{value.get('countryCode')}",
            national_number=value.get("nationalNumber"),
        )



from __future__ import annotations
from typing import Optional, TYPE_CHECKING, Tuple, Dict, Union
import traceback, flask_sock
from simple_websocket.ws import ConnectionClosed

from ..contract import should_send, common
from ..utils import deserialize, serialize, decode_jwt
from .execution import Execution, RequestData, NoExecutionFound

if TYPE_CHECKING:
    from ..repositories.json.classes import DashJSON, FormJSON


def get_live_execution_throwable() -> LiveExecution:
    execution = LiveExecution.get_execution()
    if not execution:
        raise NoExecutionFound
    return execution


class AuthResponse:
    """The response from the authentication process

    Attributes:
      email (str): The email address of the user
    """

    def __init__(self, email: str):
        self.email = email


class LiveExecution(Execution):
    type = "execution"
    _connection: flask_sock.Server

    @staticmethod
    def get_execution() -> Optional["LiveExecution"]:
        execution = Execution.get_execution()
        if isinstance(execution, LiveExecution):
            return execution

        return None

    @staticmethod
    def broadcast(msg: common.Message):
        for e in list(Execution.executions.values()):
            if isinstance(e, LiveExecution) and e.connected and e.closed == False:
                e.send(msg)

    def __init__(
        self,
        runtime_json: Union["FormJSON", "DashJSON"],
        is_initial: bool,
        connection: flask_sock.Server,
        request: RequestData,
        execution_id=None,
    ):
        self._connection = connection
        if execution_id is not None:
            self.id = execution_id

        super().__init__(runtime_json, is_initial, request, execution_id=execution_id)

    def send(self, msg: common.Message):
        self.log(msg.type, msg.data)

        if not should_send(msg, self.is_preview):
            return

        str_data = serialize(msg.to_json(self.is_preview))
        self._connection.send(str_data)

    def recv(self) -> Tuple[str, Dict]:
        try:
            str_data = self._connection.receive()
        except ConnectionClosed:
            self.close()
            return "close", {}

        data = deserialize(str_data)
        self.log(data["type"], data)
        return data["type"], data

    def close(self, reason: Optional[str] = ""):
        self._connection.close(message=reason)
        self.closed = True

    def stdio(self, type: str, text: str):
        self.send(common.StdioMessage(type, text))
        return super().stdio(type, text)

    @property
    def connected(self) -> bool:
        return self._connection.connected

    def end(self):
        if self.connected:
            self.close(reason=traceback.format_exc() or None)

    def handle_lock_failed(self):
        status = self.stage_run.status if self.stage_run else None
        self.send(common.LockFailedMessage(status))
        return super().handle_lock_failed()

    # flows

    @property
    def query_params(self) -> Dict:
        return self.context.get("query_params", {})

    @query_params.setter
    def query_params(self, value: Dict) -> None:
        self.context["query_params"] = value

    def receive(self):
        while True:
            type, data = self.recv()
            if type in ["heartbeat", "browser:try-disconnect"]:
                continue

            return data

    def get_user(self, refresh: bool = False):
        self.send(common.AuthRequireInfoMessage(refresh=refresh))

        while True:
            type, data = self.recv()
            if type != "auth:saved-jwt":
                continue

            jwt_claims = decode_jwt(data["jwt"])
            if not jwt_claims:
                self.send(common.AuthInvalidJWTMessage())
                continue

            self.send(common.AuthValidJWTMessage())
            return AuthResponse(jwt_claims["email"])

    def execute_js(self, code: str, context: dict = {}):
        self.send(common.ExecuteJSRequestMessage(code, context))

        while True:
            type, data = self.recv()
            if type != "execute-js:response":
                continue

            return data.get("value")

    def alert(self, message: str, severity: str):
        severity = (
            severity if severity in ["info", "warn", "error", "success"] else "info"
        )
        self.send(common.AlertMessage(message, severity))

    def redirect(self, url: str, query_params: dict):
        query_params = query_params or self.query_params
        self.send(common.RedirectMessage(url, query_params))



import traceback
from abc import ABC
from typing import Any, Dict, List, Optional, Tuple, Union

from ...contract import dashes_contract as contract
from ...repositories.json.classes import (
    DashJSON,
    DashWidgetJSON,
    SlotJSON,
    SlottableJSON,
)
from ...utils import formated_traceback_error_message
from ...widgets import (
    Input,
    Output,
    get_widget_class,
    input_types,
    is_broker_prop_form_only,
)
from ..live_execution import LiveExecution
from .autocomplete import get_suggestions
from .program import PythonProgram


def preview_only_check(method):
    def wrapper(self: "DashRuntime", *args, **kwargs):
        if self.execution.is_preview:
            method(self, *args, **kwargs)
            return

        self.execution.send(
            contract.ErrorMessage("This method is only allowed in preview mode")
        )

    return wrapper


class IfSlottable:
    visible: bool

    def __init__(self, metadata: SlottableJSON) -> None:
        self.metadata = metadata
        self.visible = True

    def set_props(self, props: dict):
        self.visible = props["condition"]

    @property
    def props(self):
        return self.metadata.props.items()

    def render(self, context: dict):
        return dict(condition=self.visible)


class ItemRuntime(ABC):
    props_errors: Dict[str, str]
    runtime_state: Optional[Union[IfSlottable, Input, Output]]
    props: List[Tuple[str, str]]
    metadata: Union[DashWidgetJSON, SlottableJSON]

    def set_props(self, props: dict):
        raise NotImplementedError("set_props not implemented")


def iter_children(slot: SlotJSON):
    for child_id, child_metadata in slot.items():
        if isinstance(child_metadata, DashWidgetJSON):
            yield WidgetRuntime(child_metadata)
        elif isinstance(child_metadata, SlottableJSON):
            yield SlottableRuntime(child_metadata)
        else:
            raise Exception(f"Invalid slottable type {type(child_metadata)}")


class SlottableRuntime(ItemRuntime):
    metadata: SlottableJSON
    runtime_state: Optional[IfSlottable]
    _children: List[ItemRuntime]

    def __init__(self, metadata: SlottableJSON):
        self.metadata = metadata
        self.runtime_state = None
        self._children = []
        self.props_errors = {}

    def set_props(self, props: dict):
        if not self.runtime_state:
            self.runtime_state = IfSlottable(self.metadata)
        self.runtime_state.set_props(props)
        self._children = list(iter_children(self.metadata.slot))

    @property
    def props(self):
        return self.metadata.props.items()

    @property
    def children(self):
        if self.runtime_state and self.runtime_state.visible:
            return self._children
        else:
            return []

    def make_children(self):
        self._children = list(iter_children(self.metadata.slot))

    def flatten_children(self):
        yield self
        for child in self.children:
            if not child.runtime_state:
                continue
            if isinstance(child, SlottableRuntime):
                yield from child.flatten_children()
            else:
                yield child


class RootRuntime:
    slot: SlotJSON
    children: List[ItemRuntime]

    def __init__(self, slot: SlotJSON):
        self.slot = slot
        self.children = list(iter_children(slot))

    def update(self, slot: SlotJSON):
        self.slot = slot
        self.children = list(iter_children(slot))

    def flatten_children(self):
        for child in self.children:
            if not child.runtime_state:
                continue
            if isinstance(child, SlottableRuntime):
                yield from child.flatten_children()
            else:
                yield child


class WidgetRuntime(ItemRuntime):
    runtime_state: Optional[Union[Input, Output]]
    metadata: DashWidgetJSON
    variable_error: Optional[Exception]

    def __init__(self, metadata: DashWidgetJSON):
        self.metadata = metadata
        self.runtime_state = None
        self.variable_error = None
        self.props_errors = {}

    @property
    def props(self):
        return filter(
            lambda prop: not is_broker_prop_form_only(self.metadata.type, prop[0]),
            self.metadata.props.items(),
        )

    def set_props(self, props: dict):
        if self.runtime_state:
            self.runtime_state.set_props(props)
        else:
            self.runtime_state = get_widget_class(self.metadata.type)(**props)


class DashRuntime:
    py: PythonProgram
    execution: LiveExecution
    root_slot_runtime: RootRuntime
    dash_json: DashJSON
    seq: int
    code: str

    def __init__(self, execution: LiveExecution, dash_json: DashJSON) -> None:
        self.code = dash_json.file_path.read_text(encoding="utf-8")
        self.py = PythonProgram(self.code)
        self.execution = execution
        self.root_slot_runtime = RootRuntime(dash_json.layout.slot)
        self.dash_json = dash_json
        self.seq = 0

    def handle(self, type: str, data):
        handlers = {
            "client-start": self.start,
            "widget-input": self.widget_input,
            "widget-event": self.widget_event,
            "browser:try-disconnect": self.log_only,
            "refresh:request": self.refresh_request,
            "eval": self.eval,
            "widgets-changed": self.widgets_changed,
            "variable-created": self.variable_created,
            "autocomplete:load": self.autocomplete_load,
        }
        handler = handlers.get(type, self.default_handler)
        handler(data)

    def default_handler(self, _data):
        self.execution.send(contract.ErrorMessage("unknown type"))

    def start(self, data):
        # data: { type: client-start, params: PARAMS }
        self.execution.query_params = data.get("params", {})
        try:
            self.py.execute_initial_code()
            self._compute_and_send_widgets_props()
        except Exception as e:
            tb = traceback.extract_tb(e.__traceback__)
            formated_error = formated_traceback_error_message(e, self.dash_json.file)

            self.execution.send(contract.ProgramStartFailedMessage(formated_error, tb))
            self.execution.close()

    def widget_input(self, data):
        # data: { type: widget-input, widgetId: string, newValue: any, seq: number }
        self.seq = data["seq"]
        if self.execute_widget_input(data["widgetId"], data["newValue"]):
            self.execute_widget_event(
                data["widgetId"], "update:value", data["newValue"]
            )
            self._compute_and_send_widgets_props()

    def widget_event(self, data):
        # data: { type: widget-event, widgetId: string, event: { type: string, payload: any }, seq: number }
        self.seq = data["seq"]
        if self.execute_widget_event(
            data["widgetId"], data["event"]["type"], data["event"].get("payload", {})
        ):
            self._compute_and_send_widgets_props()

    def refresh_request(self, data):
        # data: { type: refresh:request }
        self._compute_and_send_widgets_props()

    def log_only(self, _data):
        # data: { type: 'browser:try-disconnect', ... }
        pass

    @preview_only_check
    def eval(self, data):
        # data: {type: eval, expression: string}
        try:
            try:
                value = self.py.ev(data["expression"])
                self.execution.send(contract.EvalReturnMessage(repr(value)))
            except SyntaxError:
                self.py.ex(data["expression"])
                self.execution.send(contract.EvalReturnMessage(""))
        except Exception as e:
            tb = traceback.extract_tb(e.__traceback__)
            formated_error = formated_traceback_error_message(e, self.dash_json.file)
            self.execution.send(contract.EvalErrorMessage(formated_error))

        self._compute_and_send_widgets_props()

    @preview_only_check
    def widgets_changed(self, data):
        # data: { type: widgets-changed, dashDefinition: { slot }, state }
        self.root_slot_runtime.update(SlotJSON(data["dashDefinition"]["slot"]))
        self.py.root = self.root_slot_runtime.slot
        self._compute_and_send_widgets_props()

    @preview_only_check
    def variable_created(self, data):
        # data: { type: variable-created, name, value?, state }
        self.py.set_variable(data["name"], data.get("value"))
        self._compute_and_send_widgets_props()

    @preview_only_check
    def autocomplete_load(self, data):
        # data: { type: autocomplete:load, suggestionsFor: string, code: string }
        try:
            suggestions = self.get_autocomplete_suggestions(data["code"])
        except Exception as e:
            suggestions = []

        self.execution.send(
            contract.AutocompleteSuggestionsMessage(
                suggestions=suggestions,
                suggestionsFor=data["suggestionsFor"],
            )
        )

    def make_widgets_computed_response(self) -> contract.WidgetsComputedMessage:
        props = {}
        variables = {}
        errors = {}
        for child in self.root_slot_runtime.flatten_children():
            if not child.runtime_state:
                raise Exception("Widget not computed")
            props[child.metadata.id] = child.runtime_state.render({})
            if isinstance(child.runtime_state, Input):
                variables[child.metadata.id] = child.runtime_state.serialize_value()
            errors[child.metadata.id] = child.props_errors

        return contract.WidgetsComputedMessage(
            props=props,
            variables=variables,
            errors=errors,
            seq=self.seq,
        )

    def _compute_and_send_widgets_props(self):
        try:
            self.evaluate_widgets()
            self.execution.send(self.make_widgets_computed_response())
        except Exception as e:
            tb = traceback.extract_tb(e.__traceback__)
            formated_error = formated_traceback_error_message(e, self.dash_json.file)
            traceback.print_exc()
            self.execution.send(
                contract.WidgetsComputedGeneralErrorMessage(formated_error)
            )

    def execute_widget_event(self, wid, type, payload):
        widget = self.__get_widget(wid)
        cmd = widget.metadata.events.get(type)

        if not cmd:
            return False

        self.__execute_widget_event_command(widget, cmd, payload)
        return True

    def __execute_widget_event_command(self, widget: WidgetRuntime, cmd: str, payload):
        # TODO: use widget runtime to parse event payload
        self.py.state.update({"__event__": payload})

        try:
            self.py.ex(cmd)
        except Exception as e:
            traceback.print_exc()
            return {"repr": traceback.format_exc()}
        finally:
            self.py.state.pop("__widget__", None)
            self.py.state.pop("__event__", None)

    def execute_widget_input(self, wid: str, new_value: Any):
        widget = self.__get_widget(wid)
        if not widget.metadata.variable:
            return False

        self.py.set_variable(widget.metadata.variable, new_value)
        return True

    def evaluate_widgets(self):
        self.__compute_root()

    def __compute_root(self):
        for child in self.root_slot_runtime.children:
            if isinstance(child, SlottableRuntime):
                self.__compute_slottable_widgets(child)
            elif isinstance(child, WidgetRuntime):
                self.__compute_widget(child)
            else:
                raise Exception(f"Invalid slot type {type(child)}")

    def __compute_slottable_widgets(self, slot: Optional[SlottableRuntime]):
        if not slot or not isinstance(slot, SlottableRuntime):
            raise Exception(f"Invalid slot type: {type(slot)}")
        self.__compute_slot_props(slot)
        for child in slot.children:
            if isinstance(child, SlottableRuntime):
                self.__compute_slottable_widgets(child)
            elif isinstance(child, WidgetRuntime):
                self.__compute_widget(child)
            else:
                raise Exception(f"Invalid slot type {type(slot)}")

    def __compute_widget(self, widget: WidgetRuntime):
        try:
            self.__compute_slot_props(widget)
            if isinstance(widget.runtime_state, Input):
                self.__compute_widget_variable(widget)
        except Exception as e:
            traceback.print_exc()
            widget.props_errors["__general__"] = str(e)

    def __compute_slot_props(self, slot: ItemRuntime):
        if slot.metadata.type in [i.type for i in input_types] and issubclass(
            get_widget_class(slot.metadata.type), Input
        ):
            props = dict(key=None)
        else:
            props = {}
        for prop, expr in slot.props:
            try:
                props[prop] = self.py.ev(expr) if expr else None
            except Exception as e:
                traceback.print_exc()
                slot.props_errors[prop] = str(e)
        slot.set_props(props)

    def __compute_widget_variable(self, widget: WidgetRuntime):
        if not widget.runtime_state:
            raise Exception("Widget not computed")
        if isinstance(widget.runtime_state, Output):
            raise Exception("Output widgets can't have variables")

        if not widget.metadata.variable:
            parsed_value = widget.runtime_state.parse_value(None)
            widget.runtime_state.set_value(parsed_value)
            return

        try:
            self.py.ev(widget.metadata.variable)
            self.py.ex(f"{widget.metadata.variable} = {widget.metadata.variable}")
            variable_value = self.py.ev(widget.metadata.variable)
            parsed_value = widget.runtime_state.parse_value(variable_value)
            widget.runtime_state.set_value(parsed_value)
        except Exception as e:
            traceback.print_exc()
            widget.variable_error = e

    def __get_widget(self, wid: str):
        widget = search_widget_runtime(self.root_slot_runtime, wid)
        if not widget:
            raise Exception(f"Widget {wid} not found")
        return widget

    def get_autocomplete_suggestions(self, added_code_snippet):
        return get_suggestions(self.code, added_code_snippet)


def search_widget_runtime(
    slot: Union[RootRuntime, SlottableRuntime], wid: str
) -> Optional[WidgetRuntime]:
    for child in slot.children:
        if isinstance(child, WidgetRuntime) and child.metadata.id == wid:
            return child
        elif isinstance(child, SlottableRuntime):
            widget = search_widget_runtime(child, wid)
            if widget:
                return widget
    return None


def merge_errors(errors, errors_slot):
    errors["props"] = {**errors["props"], **errors_slot["props"]}
    errors["variables"] = {**errors["variables"], **errors_slot["variables"]}
    errors["widgets"] = {**errors["widgets"], **errors_slot["widgets"]}



import threading, os, time
from datetime import datetime
from pathlib import Path

from ..execution.live_execution import LiveExecution
from ..contract.common import FilesChangedMessage
from ..modules import reload_modules_from_path
from ..settings import Settings


def __files_changed_polling_loop(path: Path):
    last_change = datetime.now().timestamp()
    while True:
        something_changed = False
        for _, _, files in list(os.walk(path)):
            for file in files:
                if not Path(file).exists() or Path(file).name == "abstra.json":
                    continue
                st_mtime = os.stat(file).st_mtime
                if st_mtime > last_change:
                    last_change = st_mtime
                    something_changed = True

        if something_changed:
            reload_modules_from_path(path)
            LiveExecution.broadcast(FilesChangedMessage())

        time.sleep(1)


def watch_py_root_files():
    threading.Thread(
        target=lambda: __files_changed_polling_loop(Settings.root_path)
    ).start()



import uuid, sys, importlib, importlib.util as imp_util
from pathlib import Path


def import_as_new(name: str):
    module_spec = imp_util.spec_from_file_location(uuid.uuid4().__str__(), name)
    if module_spec is None or module_spec.loader is None:
        raise Exception(f"Could not load module {name}")

    module = imp_util.module_from_spec(module_spec)
    if module is None:
        raise Exception(f"Could not load module {name}")

    module_spec.loader.exec_module(module)


# TODO: make this work with virtualenv
def reload_modules_from_path(path: Path):
    for module in list(sys.modules.values()):
        if not (hasattr(module, "__file__") and module.__file__):
            continue

        module_path = Path(module.__file__)
        if not module_path.is_file():
            continue

        if module.__file__.startswith(str(path)):
            importlib.reload(module)



import requests, typing, concurrent.futures as futures, dataclasses

from .utils.environment import LOG_SERVICE_URL

executor = futures.ThreadPoolExecutor()


@dataclasses.dataclass
class LogMessage:
    event: str
    executionId: str
    payload: typing.Dict[str, str]
    runtime_type: str
    runtime_name: str

    def json(self):
        return dict(
            event=str(self.event),
            executionId=str(self.executionId),
            payload=dict(self.payload),
            runtimeType=str(self.runtime_type),
            runtimeName=str(self.runtime_name),
        )


def async_log(msg: LogMessage):
    if not LOG_SERVICE_URL:
        return
    try:
        requests.post(LOG_SERVICE_URL, json=msg.json())
    except Exception as e:
        pass  # TODO: handle this


def log(msg: LogMessage):
    if not LOG_SERVICE_URL:
        return
    executor.submit(async_log, msg)



import sys
from .execution.execution import Execution

_og_stdout_write = sys.stdout.write
_og_stderr_write = sys.stderr.write


def writeWraper(type, write, text: str):
    try:
        write(text)
        execution = Execution.get_execution()
        if execution and text.strip() != "":
            execution.stdio(type, text)
    finally:
        return len(text)


def overloads():
    sys.stdout.write = lambda text: writeWraper("stdout", _og_stdout_write, text)
    sys.stderr.write = lambda text: writeWraper("stderr", _og_stderr_write, text)


def reset():
    sys.stdout.write = _og_stdout_write
    sys.stderr.write = _og_stderr_write



import warnings


def strict_compatible_missing_entities(data):
    if "hooks" not in data:
        data["hooks"] = []

    if "jobs" not in data:
        data["jobs"] = []

    if "forms" not in data:
        data["forms"] = []

    if "dashes" not in data:
        data["dashes"] = []

    if "scripts" not in data:
        data["scripts"] = []

    if "workspace" not in data:
        data["workspace"] = {
            "name": "Untitled Workspace",
            "sidebar": [],
        }

    if "workflow_transitions" not in data:
        data["workflow_transitions"] = []

    return data


def strict_compatible_workflow_positions(data: dict):
    # On release v{TBD} we added workflow positions to the abstra.json file.
    # By default, positions should be placed in columns
    size = 110
    forms = data["forms"]
    for idx, form in enumerate(forms):
        if "workflow_position" not in form:
            form["workflow_position"] = [0, idx * size]

    for idx, job in enumerate(data["jobs"]):
        if "workflow_position" not in job:
            job["workflow_position"] = [size, idx * size]

    for idx, hook in enumerate(data["hooks"]):
        if "workflow_position" not in hook:
            hook["workflow_position"] = [2 * size, idx * size]
    return data


def strict_compatible_is_initial(data: dict):
    for form in data["forms"]:
        if "is_initial" in form:
            del form["is_initial"]

    for job in data["jobs"]:
        if "is_initial" in job:
            del job["is_initial"]

    for hook in data["hooks"]:
        if "is_initial" in hook:
            del hook["is_initial"]

    for script in data["scripts"]:
        if "is_initial" in script:
            del script["is_initial"]

    return data


def strict_compatible_transitions(data: dict):
    for form in data["forms"]:
        if "transitions" not in form:
            form["transitions"] = []

    for job in data["jobs"]:
        if "transitions" not in job:
            job["transitions"] = []

    for hook in data["hooks"]:
        if "transitions" not in hook:
            hook["transitions"] = []
    return data


def _find_transition_target(data: dict, target_path: str, target_type: str, warn: bool):
    for form in data["forms"]:
        if form["path"] == target_path and target_type == "forms":
            return form

    for job in data["jobs"]:
        if job["identifier"] == target_path and target_type == "jobs":
            return job

    for hook in data["hooks"]:
        if hook["path"] == target_path and target_type == "hooks":
            return hook

    for script in data["scripts"]:
        if script["path"] == target_path and target_type == "scripts":
            return script

    if warn:
        warnings.warn(
            f"Could not find transition target {target_type}:{target_path}. This will be ignored, but please, make sure to remove it from your abstra.json file"
        )
    return None


def strict_compatible_remove_dangling_transitions(data: dict):
    for form in data["forms"]:
        form["transitions"] = [
            transition
            for transition in form["transitions"]
            if _find_transition_target(
                data, transition["target_path"], transition["target_type"], warn=True
            )
            is not None
        ]

    for job in data["jobs"]:
        job["transitions"] = [
            transition
            for transition in job["transitions"]
            if _find_transition_target(
                data, transition["target_path"], transition["target_type"], warn=True
            )
            is not None
        ]

    for hook in data["hooks"]:
        hook["transitions"] = [
            transition
            for transition in hook["transitions"]
            if _find_transition_target(
                data, transition["target_path"], transition["target_type"], warn=True
            )
            is not None
        ]

    return data


def strict_compatible(data: dict):
    data = strict_compatible_missing_entities(data)
    data = strict_compatible_workflow_positions(data)
    data = strict_compatible_is_initial(data)
    data = strict_compatible_transitions(data)
    data = strict_compatible_remove_dangling_transitions(data)
    return data
